require 'libs.util'

MessageController = util.singleton( "MessageController" )
MessageController.turn = 0
MessageController.timeElapsedSinceBeginningOfTurn = 0
MessageController.state = 0
MessageController.numPeers = 0
MessageController.numPeersCompletedTurn = 0
MessageController.playerid = 0
MessageController.messageQueue = {}

MessageController.TURN_DURATION                 = 0.08
MessageController.STATE_IN_TURN                 = 1
MessageController.STATE_WAITING_FOR_NEXT_TURN   = 2

function MessageController:sendMessage( messageType, message )
    local message = message or {}
    message['_m'] = messageType
    self.messageQueue[#self.messageQueue+1] = message
end

function MessageController:update( dt )
    self.timeElapsedSinceBeginningOfTurn = self.timeElapsedSinceBeginningOfTurn + dt
    if self.state == self.STATE_IN_TURN and self.timeElapsedSinceBeginningOfTurn >= self.TURN_DURATION then
        self:completeTurn()
    end
    
    if self.state == self.STATE_WAITING_FOR_NEXT_TURN and self.numPeersCompletedTurn > 0 and self.numPeersCompletedTurn == self.numPeers then
        self:nextTurn()
    end
end

function MessageController:completeTurn()
    self.state = self.STATE_WAITING_FOR_NEXT_TURN
    self:sendMessage( MSG_CLIENT_TURN_COMPLETE )
    self:sendMessageQueue()
    self.messageQueue = {}
end

function MessageController:nextTurn()
    self.turn = self.turn + 1
    self.state = self.STATE_IN_TURN
    self.numPeersCompletedTurn = 0
    self.timeElapsedSinceBeginningOfTurn = 0
end

function MessageController:sendMessageQueue()
    self.messageQueue['_m'] = MSG_QUEUE
    self.messageQueue['_p'] = self.playerid
    self.messageQueue['_t'] = self.turn
    NetworkController:sendTableToServer( self.messageQueue )
end

function MessageController:onReceiveMessageQueue( messageQueue, turn, playerid )
    if turn == self.turn then
        for i, message in ipairs( messageQueue ) do
            local messageid = message['_m']
            self.readMessage[ messageid ]( unpack(message) )
        end
    end
end

function MessageController:onReceiveMessage( message )
    local messageid = message['_m']
    if messageid == MSG_QUEUE then
        self:onReceiveMessageQueue( message, message['_t'], message['_p'] )
    else
        self.readMessage[ messageid ]( unpack(message) )
    end
end

function MessageController:haveAllPeersCompletedTurn()
    return self.numPeersCompletedTurn == self.numPeers
end

MSG_QUEUE                       = 0
MSG_CLIENT_TURN_COMPLETE        = 1
MSG_CLIENT_CONNECTED            = 2
MSG_CLIENT_DISCONNECTED         = 3
MSG_SNAPSHOT                    = 4
MSG_PROGRAM_ORDER_MOVE          = 101
MSG_PROGRAM_ORDER_ATTACK        = 102
MSG_PROGRAM_ORDER_ASSIST        = 103
MSG_PROGRAM_ORDER_GOAL_PORT     = 104
MSG_PROGRAM_ORDER_GOAL_SYSTEM   = 105
MSG_PORT_FORWARD                = 201

local op = {}

op[ MSG_QUEUE ] = function( ... )
    MessageController.onReceiveMessageQueue( MessageController, {...} )
end

op[ MSG_SNAPSHOT ] = function( playerid, turn, numPeers, seed )
    RNG = twister( seed )
    GridController():generateMap()
    
    MessageController.playerid = playerid
    MessageController.turn = turn
    MessageController.numPeers = numPeers
    NetworkController:receivedPlayerID( playerid )
end

op[ MSG_CLIENT_TURN_COMPLETE ] = function()
    MessageController.numPeersCompletedTurn = MessageController.numPeersCompletedTurn + 1
    --print( MessageController.numPeersCompletedTurn .. " / " .. MessageController.numPeers .. " peers completed turn." )
end

op[ MSG_CLIENT_CONNECTED ] = function()
    print("Peer connected.")
    MessageController.numPeers = MessageController.numPeers + 1
    print("Number of peers: "..MessageController.numPeers)
end

op[ MSG_CLIENT_DISCONNECTED ] = function()
    print("Peer disconnected.")
    MessageController.numPeers = MessageController.numPeers - 1
end

op[ MSG_PROGRAM_ORDER_MOVE ] = function( programid, x, y )
    local program = EntityController.getEntityByID( programid )
    if program then
        program:addMoveOrder( x, y )
    end
end

op[ MSG_PROGRAM_ORDER_GOAL_PORT ] = function( programid, portid )
    local program   = EntityController.getEntityByID( programid )
    local port      = EntityController.getEntityByID( portid )
    if program and port then
        program:addGoalPortOrder( port )
    end
end

op[ MSG_PROGRAM_ORDER_GOAL_SYSTEM ] = function( programid, systemid )
    local program   = EntityController.getEntityByID( programid )
    local system    = EntityController.getEntityByID( systemid )
    if program and system then
        program:addGoalSystemOrder( system )
    end
end

op[ MSG_PORT_FORWARD ] = function( portAid, portBid )
    local portA     = EntityController.getEntityByID( portAid )
    local portB     = EntityController.getEntityByID( portBid )
    if portA and portB then
        portA:forwardToPort( portB )
    end
end

op[ MSG_PROGRAM_ORDER_ATTACK ] = function( programid, targetid )
    local program   = EntityController.getEntityByID( programid )
    local target    = EntityController.getEntityByID( targetid )
    if program and target then
        program:addAttackOrder( target )
    end
end

op[ MSG_PROGRAM_ORDER_ASSIST ] = function() end

MessageController.readMessage = op

