PortController = class('PortController', EntityController)
PortController.instances = {}
-- requires
require "PortInputController"
require "PortRenderer"
-- States
PORT_UNCONNECTED    = 0
PORT_CONNECTED      = 1
PORT_OPEN           = 2
PORT_FORWARDED      = 3
--
-- upgrades
PORT_UPGRADE_NONE       = 0
PORT_UPGRADE_FIREWALL   = 1
--

function PortController:initialize( systemController, side, portRenderer )
    EntityController.initialize( self )
	self.systemController = systemController
	self.systemControllerSide = side
	self.packets = {}
	self.transferSpeed = 70
	self.forwardSpeed = 5
	self.upgrade = PORT_UPGRADE_NONE
	self.upgradeProperties = nil
	self.upgradable = true
	self.state = PORT_UNCONNECTED
	self.sisterPort = nil
	self.forwardPort = nil
	self.isMaster = false
	self.backdoor = false
	self.portRenderer = portRenderer
	self.portRenderer:setPortController( self )
	self.inputController = PortInputController( self )
	self.boundingBox = BoundingBox(0,0,25,25)
	self:calculatePosition()
	table.insert( PortController.instances, self )
end

function PortController:calculatePosition()
    local edge = self.systemController.hexagonShape:getEdgeBySide( self.systemControllerSide, 25 )
	local t = (1/6) + (1/5) * self.systemController:numberOfPortsOnSide( self.systemControllerSide )
    if self.isMaster then
        self.position.x, self.position.y = util.lerpXY( edge[1], edge[2], edge[3], edge[4], t )
    else
        local a, b, c, d = util.reverseXY(edge[1], edge[2], edge[3], edge[4])
        self.position.x, self.position.y = util.lerpXY( a, b, c, d, t )
    end
    self.boundingBox:placeAroundPosition( self.position )
end

function PortController:update( dt )
    --if true then return false end
    -- update input contoller.
    self.inputController:update( dt )
    -- update transfer progress of all packets.
    for i, packet in ipairs(self.packets) do
        packet.transferProgress = packet.transferProgress + (self.transferSpeed * dt)
        if packet.transferProgress >= 100 then
            self:deliverPacket( packet )
        end
    end
end

function PortController:connectToPort( port )
	self.sisterPort = port
	self.state = PORT_CONNECTED
	self.isMaster = true
	port.sisterPort = self
	port.state = PORT_CONNECTED
	port.isMaster = false
	self:calculatePosition()
	port:calculatePosition()
end

function PortController:canForwardToPort( port )
    return port.systemController == self.systemController 
            and port ~= self 
            and self.backdoor == false
            and port.backdoor == false
            and port:isConnected() 
            and not port:isLoopedToPort( self )
            and not self:forwardLaneIntersectsWithOtherForwardLanes( self.position, port.position )
end

function PortController:forwardLaneIntersectsWithOtherForwardLanes( a, b )
    local ports = self.systemController.ports
    for i, port in ipairs(ports) do
        if port:isForwarded() then
            local c = port.position
            local d = port.forwardPort.position
            if util.segIntersection( a.x, a.y, b.x, b.y, c.x, c.y, d.x, d.y ) then
                return true
            end
        end
    end
    return false
end

function PortController:forwardToPort( port )
    -- check port is in same system and not self.
    if self:canForwardToPort( port ) then
        self.forwardPort = port
        self.state = PORT_FORWARDED
        self.forwardMaster = true
        port.forwardPort = self
        port.state = PORT_FORWARDED
        port.forwardMaster = false
        return true
    else
        return false
    end
end

function PortController:removeForwardPort()
    if self:isForwarded() then
        self.forwardPort.forwardPort = nil
        self.forwardPort.state = PORT_CONNECTED
        self.state = PORT_CONNECTED
        self.forwardMaster = false
        self.forwardPort = nil
        return true
    else
        return false
    end
end

function PortController:canTransferProgram( program )
    if self:isConnected() then
        -- disallow enemy programs if firewall enabled.
        local programTeam = program:getTeam()
        if self.upgrade == PORT_UPGRADE_FIREWALL then
            if programTeam == self.upgradeProperties.team then
                return true
            else
                return false
            end
        end
        -- default when connected.
        return true
    end
    
    return false
end

function PortController:transferProgram( program )
    if not self:canTransferProgram( program ) then return false end
    -- add program to any existing packets that are still at 0% progress.
    -- otherwise create a new packet.
    if #self.packets > 0 and self.packets[ #self.packets ].transferProgress == 0 then
        local packet = self.packets[ #self.packets ]
        packet.programs[#packet.programs+1] = program
    else
        local packet = {
            transferProgress = 0,
            programs = { program }
        }
        self.packets[#self.packets+1] = packet
    end

    -- tell sister port about transfer
    self.sisterPort:onPortTransferPacket( self, packet )
    return true
end

function PortController:deliverPacket( packet, key )
    -- check packet exists
    if not packet then
        error( "packet == nil" )
        return false
    end
    
    -- remove packet from packets array.
    self:removePacket( packet )
    -- tell sister port about the packet.
    self.sisterPort:onPortDeliveredPacket( self, packet )
end

function PortController:removePacket( packet )
    for i, testPacket in ipairs(self.packets) do
        if testPacket == packet then
            table.remove( self.packets, i )
            return true
        end
    end
    return false
end

function PortController:onPortTransferPacket( port, packet )
end

function PortController:onPortDeliveredPacket( port, packet )
    if self.state == PORT_FORWARDED then
        self.forwardPort:onPortForwardedPacket( self, packet )
    else
        for i, program in ipairs(packet.programs) do
            self.systemController:addProgram( program )
            program:onExitPort( self, self.systemController, self.position )
        end
    end
end

function PortController:onPortForwardedPacket( port, packet )
    packet.transferProgress = 0
    self.packets[#self.packets+1] = packet
    self.sisterPort:onPortTransferPacket( self, packet )
end

function PortController:setTransferSpeed( speed )
    self.transferSpeed = speed
    if self.state == PORT_CONNECTED then
        self.sisterPort.transferSpeed = speed
    end
end

function PortController:getSisterPort()
    return self.sisterPort
end

function PortController:getForwardPort()
    return self.forwardPort
end

function PortController:getConnectedSystem()
    if self:isConnected() then
        if self.sisterPort:isForwarded() then
            local port = self.sisterPort
            while port:isForwarded() and port.forwardPort.sisterPort do
                port = port.forwardPort.sisterPort
            end
            return port.systemController
        else
            return self.sisterPort.systemController
        end
    end
end

function PortController:isConnected()
    return (self.state == PORT_CONNECTED) and self.sisterPort
end

function PortController:isForwarded()
    return (self.state == PORT_FORWARDED) and self.forwardPort
end

function PortController:isForwardMaster()
    return self.forwardMaster
end

function PortController:isLoopedToPort( port )
    local _port = self
    while _port.sisterPort and _port.sisterPort:isForwarded() do
        _port = _port.sisterPort.forwardPort
    end
    return _port.sisterPort and (_port.sisterPort == port)
end

function PortController:getPacketWithProgram( programController )
    for i, packet in ipairs(self.packets) do
        for _, program in ipairs(packet.programs) do
            if program == programController then
                return packet
            end
        end
    end
    return nil
end

function PortController:hasProgram( programController )
    for i, packet in ipairs(self.packets) do
        for _, program in ipairs(packet.programs) do
            if program == programController then
                return true
            end
        end
    end
    return false
end

function PortController:hasPacket( packet )
    for i, testPacket in ipairs(self.packets) do
        if testPacket == packet then
            return true
        end
    end
    return false
end

function PortController:setBackdoor( backdoor )
    self.backdoor = backdoor
    if backdoor then
        self.transferSpeed = self.transferSpeed / 4
    end
end

function PortController:isBackdoor()
    return self.backdoor
end