require 'EntityController'
require 'BoundingBox'
require 'ProgramInputController'
require 'ProgramRenderer'

ProgramController = class('ProgramController', EntityController)
ProgramController:include( Stateful )
ProgramController.STATE_IDLE                    = ProgramController:addState('IDLE')
ProgramController.STATE_APPROACH_PORT           = ProgramController:addState('APPROACH_PORT')
ProgramController.STATE_INSIDE_PORT             = ProgramController:addState('INSIDE_PORT')
ProgramController.STATE_NAVIGATING_SYSTEM_PATH  = ProgramController:addState('NAVIGATING_SYSTEM_PATH')

STATE_STATIONARY        = 4
STATE_INSIDE_PORT       = 5

-----------------------------
--      ORDERS
-----------------------------
ProgramController.runOrder = {}
function ProgramController.runOrder:move( order )
    self:setDestination( vector( order.x, order.y ), true )
end

function ProgramController.runOrder:port( order )
    if not order.keepGoals then self:clearGoals() end
    self:gotoState('APPROACH_PORT')
    self.goalPort = order.portController
    self:setDestination( order.portController:getPosition() )
end

function ProgramController.runOrder:system( order )
    self:gotoState('NAVIGATING_SYSTEM_PATH')
    GridController.instance.pathfinder:findPath( self.systemController, order.systemController, function(path)
        if path then
            self.pathToExternalSystem = path
            self.goalSystem = path[#path]
            self:approachPortThatLeadsToNextSystemInPath( path )
        end
    end)
end

function ProgramController.runOrder:attack( order )
    self:clearGoals()
    self.attackTarget = order.target
    if self.attackTarget.systemController then
        if self.attackTarget.systemController ~= self.systemController then
            self:addGoalSystemOrder( self.attackTarget.systemController )
        else
            self:setDestination( order.target:getPosition() )
        end
    end
end

function ProgramController.runOrder:assist( order )
end
-----------------------------

function ProgramController:initialize( inputController, systemController, programRenderer )
    EntityController.initialize( self )
    self:gotoState('IDLE')
    self.systemController = systemController
    self.position = vector( 0, 0 )
    self.boundingBox = BoundingBox(-10, -10, 10, 10)
    self.inputController = inputController
    self.inputController:setProgramController(self)
    if programRenderer then
        self.programRenderer = programRenderer
        self.programRenderer:setProgramController(self)
    end
    self.orderQueue = {}
    self.selected = false
    
    self.team = 0
    self.speed = 200
    self.attackRange = 20
    
    self.listener = NotificationController.listen( "deselect_all", nil, self.onDeselectAllPrograms, self )
end

function ProgramController:destroy()
    EntityController.destroy( self )
    if self.systemController then
        self.systemController:removeProgram( self )
    end
    self.programRenderer:destroy()
    self.inputController:destroy()
    NotificationController.killListener( self.listener )
end

function ProgramController.STATE_NAVIGATING_SYSTEM_PATH:enteredState()
    print('enter STATE_NAVIGATING_SYSTEM_PATH')
end

function ProgramController.STATE_NAVIGATING_SYSTEM_PATH:continuedState()
    print('continue STATE_NAVIGATING_SYSTEM_PATH')
end

function ProgramController.STATE_NAVIGATING_SYSTEM_PATH:exitedState()
    print('exit STATE_NAVIGATING_SYSTEM_PATH')
end

function ProgramController.STATE_APPROACH_PORT:enteredState()
    print('enter STATE_APPROACH_PORT')
end

function ProgramController.STATE_APPROACH_PORT:continuedState()
    print('continue STATE_APPROACH_PORT')
    ProgramController.STATE_APPROACH_PORT:enteredState()
end

function ProgramController.STATE_APPROACH_PORT:exitedState()
    print('exit STATE_APPROACH_PORT')
end

function ProgramController.STATE_APPROACH_PORT:update( dt )
    self:_update( dt )
    if not self.goalPort then
        self:popState('APPROACH_PORT')
    elseif vector.distance( self.position, self.goalPort.position ) < self.speed * dt then
        self:popState('APPROACH_PORT')
        self:enterPort( self.goalPort )
        self:setDestination( nil )
        self.goalPort = nil
    end
end

function ProgramController.STATE_INSIDE_PORT:enteredState()
    print('enter STATE_INSIDE_PORT')
    if self.programRenderer then
        self.programRenderer:programEnteredState( 'INSIDE_PORT' )
    end
    self.inputController:programEnteredState( 'INSIDE_PORT' )
end

function ProgramController.STATE_INSIDE_PORT:continuedState()
    print('continue STATE_INSIDE_PORT')
    ProgramController.STATE_INSIDE_PORT:enteredState()
end

function ProgramController.STATE_INSIDE_PORT:exitedState()
    print('exit STATE_INSIDE_PORT')
    if self.programRenderer then
        self.programRenderer:programExitedState( 'INSIDE_PORT' )
    end
    self.inputController:programExitedState( 'INSIDE_PORT' )
end

function ProgramController:update( dt )
    self:_update( dt )
end

function ProgramController:_update( dt )
    self.inputController:update( dt )
    
    local order = self:popFirstInOrderQueue()
    if order then
        self.runOrder[order.orderType]( self, order )
    elseif not self:getState() then
        self:gotoState('IDLE')
    end
    
    if self.attackTarget then
        if self.attackTarget.systemController and self.attackTarget.systemController == self.systemController then 
            self:setDestination( self.attackTarget:getPosition(), false )
            if vector.distance( self.position, self.attackTarget:getPosition() ) < self.attackRange then
                -- do something to damage the target.
                self.attackTarget:destroy()
                self.attackTarget = nil
            end
        end
    end
    
    if self.destination then
        local direction = vector.normalize(self.destination - self.position)
        self:moveBy( direction * dt * self.speed )
        if vector.distance( self.position, self.destination ) <= self.speed * dt then
            self.destination = nil
        end
    end
end

function ProgramController:getSystem()
    return self.systemController
end

function ProgramController:setSystem( systemController )
    self.systemController = systemController
end

function ProgramController:approachPortThatLeadsToNextSystemInPath( path )
    -- Check we've received the path and we're still in the correct state.
    if not path then return false end
    -- Check if we're inside the goal system.
    if self.systemController == self.goalSystem then
        -- remove the path and pop the state.
        self.pathToExternalSystem = nil
        self.goalSystem = nil
        return false
    end
    -- Get a list of ports that lead to any system in the path.
    local ports = self.systemController:getPortsConnectedToAnySystemInSystems( path )
    -- Remove the next system in the path from the path.
    table.remove( self.pathToExternalSystem, 1 )
    -- iterate through the list of available ports.
    if ports and #ports > 0 then
        -- move back door ports to the back.
        for i = #ports, 1, -1 do
            local port = ports[i]
            if port.backdoor then
                table.remove(ports, i)
                ports[#ports+1] = port
            end
        end 
        -- find a port we're allowed to use.
        local didFindPort = false
        for i = 1, #ports do
            if ports[i] and ports[i]:canTransferProgram( self ) then
                self:addGoalPortOrder( ports[i], true )
                didFindPort = true
                break
            end
        end
        
        return didFindPort
    else
        return false
    end
end

function ProgramController:enterPort( portController )
    if portController:transferProgram( self ) then
        self:gotoState('INSIDE_PORT')
        self.systemController:removeProgram( self )
        self.portController = portController
        self:setPosition(0,0)
        return true
    else
        return false
    end
end

function ProgramController:onExitPort( portController, systemController, portPosition )
    self:popState('INSIDE_PORT')
    
    local direction = vector.normalize( portController:getSisterPort():getPosition() - portPosition )
    local randomPos = portPosition + (direction * RNG:random(20,200)) + vector( RNG:random(-50,50), RNG:random(-50,50) )
    self:setPosition( portPosition )
    self:setDestination( randomPos )
    self.portController = nil
    self.systemController = systemController
    self:approachPortThatLeadsToNextSystemInPath( self.pathToExternalSystem )
end

function ProgramController:position()
    return self.position
end

function ProgramController:setPosition( x, y )
    if type(x) == "table" then
        self.position = x:clone()
    else
        self.position = vector( x, y )
    end
    self.boundingBox:placeAroundPosition( self.position )
end

function ProgramController:moveBy( v )
    self.position = self.position + v
    self.boundingBox:placeAroundPosition( self.position )
end

function ProgramController:getBoundingBox()
    return self.boundingBox
end

function ProgramController:getSelected()
    return self.selected
end

function ProgramController:setSelected( selected, shouldNotify )
    if selected == true and shouldNotify then
        NotificationController.send("deselect_all_programs")
    end
    self.selected = selected
end

function ProgramController:onDeselectAllPrograms()
    self.selected = false
end

function ProgramController:addMoveOrder( x, y )
    local order = {
        orderType = 'move',
        x = x, y = y
    }
    self:addToOrderQueue( order )
end

function ProgramController:addAttackOrder( target )
    local order = {
        orderType = 'attack',
        target = target
    }
    self:addToOrderQueue( order )
end

function ProgramController:addAssistOrder( target )
    local order = {
        orderType = 'assist',
        target = target
    }
    self:addToOrderQueue( order )
end

function ProgramController:addGoalPortOrder( portController, keepGoals )
    local order = {
        orderType = 'port',
        portController = portController,
        keepGoals = keepGoals
    }
    self:addToOrderQueue( order )
end

function ProgramController:addGoalSystemOrder( systemController )
    local order = {
        orderType = 'system',
        systemController = systemController
    }
    self:addToOrderQueue( order )
end

function ProgramController:addToOrderQueue( order )
    table.insert( self.orderQueue, order )
end

function ProgramController:popFirstInOrderQueue()
    local order = self.orderQueue[1]
    table.remove( self.orderQueue, 1 )
    return order
end

function ProgramController:setTeam( team )
    self.team = team
end

function ProgramController:getTeam()
    return self.team
end

function ProgramController:setDestination( destination, shouldClearGoals )
    if shouldClearGoals then
        self:clearGoals()
    end
    self.destination = destination
end

function ProgramController:clearGoals()
    self.destination = nil
    self.pathToExternalSystem = nil
    self.goalSystem = nil
    self.goalPort = nil
    self.attackTarget = nil
end

function ProgramController:isEnemyWith( programController )
    return true
    --return self.team ~= programController:getTeam()
end

function ProgramController:isAlliedWith( programController )
    return false
    --return self.team == programController:getTeam()
end