require "Hexagon"
require "NotificationController"
require "NetworkController"
require "MessageController"
require "EntityController"
require "EntityRender"
require "GridController"
require "SystemController"
require "PortController"
require "ProgramController"

StateGame = Gamestate.new()

function StateGame:enter( oldState, isHost, ip, playername )
    
    if isHost then
        -- Generate random seed, based on current time.
        -- this seed will be distributed to clients
        RNG = twister()
        
        NetworkController:startServer()
        NetworkController:startClient( "127.0.0.1", "Host" )
    else
        NetworkController:startClient( ip, playername )
    end
    
    self.fontDebugInformation = love.graphics.newFont( 11 )
    self.fontProgramCounter   = love.graphics.newFont( FONTS_PATH .. "Agency.ttf", 14 )
    
    MessageController:nextTurn()
    hexagon = Hexagon( 100 )
    camera1 = camera( vector(16500,11200), 1/32 )
    camera1:setZoomLimits( 1/32, 1 )
    self.currentCamera = camera1
    minScale = 1/32
    maxScale = 1
    linear_steps = 5
    maxLinearScale = math.log( 1 / minScale )
    minLinearScale = math.log( 1 / maxScale )
    linear_step = (maxLinearScale - minLinearScale) / linear_steps
    linear_scale = maxLinearScale
    linear_scale_time = 1
    linear_scale_lerp_speed = 1
    target_linear_scale = linear_scale
end

function StateGame:update( dt )
    NetworkController:update( dt )
	MessageController:update( dt )
	self.currentCamera:update( dt )
	self:tweenCameraToTargetScale( dt )
	if not self.isPanning then self:tweenCameraToCenter( dt ) end
	EntityController.updateInstances( dt )
end

function StateGame:tweenCameraToTargetScale( dt )
    if target_linear_scale and linear_scale_time < 1 then
        linear_scale_time = linear_scale_time + (dt * linear_scale_lerp_speed)
        linear_scale = util.lerp( linear_scale, target_linear_scale, linear_scale_time )
        self.currentCamera:zoomToCursor( 1/math.exp(linear_scale) )
    end
end

function StateGame:tweenCameraToCenter( dt )
    if self.currentCamera.zoom <= minScale * 1.2 and tweenToCenter_timer then
        tweenToCenter_timer = tweenToCenter_timer + (dt*4)
        local x, y = util.lerpXY( camera1.pos.x, camera1.pos.y, 16500, 11200, dt )
        camera1.pos.x = x
        camera1.pos.y = y
    else
        tweenToCenter_timer = 0
    end
end

function getCurrentCamera()
    return StateGame.currentCamera
end

function StateGame:draw()
    love.graphics.setBackgroundColor( 15, 15, 15 )
    love.graphics.setColor(255,255,255)
	
	-- render all entities.
	self.currentCamera:attach()
	EntityRender.renderInstances()
	self.currentCamera:detach()
	
	-- draw the amount of programs inside a system over each system.
	if self.currentCamera.zoom < 1/10 and GridController():hasGeneratedGrid() then
    	love.graphics.setFont( self.fontProgramCounter )
    	love.graphics.setColor(0,0,0)
    	for i, system in ipairs(GridController().systems) do
    	    if #system.programs > 0 then
                local pos = self.currentCamera:cameraCoords( system.position )
                love.graphics.print(#system.programs, pos.x, pos.y)
            end
    	end
    	love.graphics.setColor(255,255,255)
	end
	
	-- print general information in the top left.
	local s = ("%s. %s systems, %s ports, %s programs. %s fps"):format(
	    NetworkController.server and "Hosting session" or "Client",
	    #GridController().systems, 
	    #PortController.instances, 
	    #EntityController.getEntitiesByClass( ProgramController ),
	    love.timer.getFPS()
	)
	love.graphics.setColor(255,255,255)
	love.graphics.setFont( self.fontDebugInformation )
	love.graphics.print( s, 10, 5 )
	
	loveframes.draw()
end

function StateGame:mousepressed( x, y, button )
    NotificationController.send("on_mouse_button_pressed", nil, x, y, button)
    if button == 'r' then
        camera1:startPan()
        self.isPanning = true
    elseif button == 'wu' then -- Zoom in
	    target_linear_scale = target_linear_scale - linear_step
	    linear_scale_time = 0
	    if target_linear_scale < minLinearScale then target_linear_scale = minLinearScale end
    elseif button == 'wd' then -- Zoom out
        target_linear_scale = target_linear_scale + linear_step
        linear_scale_time = 0
	    if target_linear_scale > maxLinearScale then target_linear_scale = maxLinearScale end
    end
end

function StateGame:mousereleased( x, y, button )
    NotificationController.send("on_mouse_button_released", nil, x, y, button)
    if button == 'r' then
	    camera1:endPan()
	    self.isPanning = false
    end
end

function StateGame:keypressed( key, unicode )
	NotificationController.send("on_key_pressed", nil, key, unicode)
	loveframes.keypressed(key,unicode)
end

function StateGame:keyreleased( key, unicode )
	NotificationController.send("on_key_released", nil, key, unicode)
	loveframes.keyreleased(key,unicode)
end