require 'specs.init'
require 'PortController'
require 'Hexagon'

describe["A Port"] = function()
    before = function()
        MockSystemControllerA = Mock:new()
        MockSystemControllerA.numberOfPortsOnSide:returns( 1 )
        MockSystemControllerA.hexagonShape = Hexagon( 100 )
        MockSystemControllerB = Mock:new()
        MockSystemControllerB.numberOfPortsOnSide:returns( 1 )
        MockSystemControllerB.hexagonShape = Hexagon( 100 )
        MockPortRenderer = Mock:new()
        MockProgramController = Mock:new()
        portA = PortController( MockSystemControllerA, DIR_N, MockPortRenderer )
        portB = PortController( MockSystemControllerB, DIR_S, MockPortRenderer )
        portA:connectToPort( portB )
    end
    
    it["should have a sister port"] = function()
        expect( portA:getSisterPort() ).should_be( portB )
    end
    
    describe["when a program enters"] = function()
        before = function()
            portA:transferProgram( MockProgramController )
        end
    
        it["should have a reference to the program"] = function()
            expect( portA:hasProgram( MockProgramController ) ).should_be( true )
        end
    end
    
    describe["when multiple programs enter"] = function()
        it["should have a single packet containing references to the programs"] = function()
            program1 = Mock:new()
            program2 = Mock:new()
            portA:transferProgram( program1 )
            portA:transferProgram( program2 )
            local packet1 = portA:getPacketWithProgram( program1 )
            local packet2 = portA:getPacketWithProgram( program2 )
            expect( packet1 ).should_not_be( nil )
            expect( packet2 ).should_not_be( nil )
            expect( packet1 ).should_be( packet2 )
            expect( #portA.packets ).should_be( 1 )
        end
    end
    
    describe["when multiple packets enter at different times"] = function()
        it["should have two different packets"] = function()
            program1 = Mock:new()
            program2 = Mock:new()
            portA.inputController = Mock:new()
            portA.transferSpeed = 1
            portA:transferProgram( program1 )
            portA:update( 1 )
            portA:transferProgram( program2 )
            local packet1 = portA:getPacketWithProgram( program1 )
            local packet2 = portA:getPacketWithProgram( program2 )
            expect( packet1 ).should_not_be( nil )
            expect( packet2 ).should_not_be( nil )
            expect( packet1 ).should_not_be( packet2 )
            expect( #portA.packets ).should_be( 2 )
        end
    end
    
    describe["when packet transfer complete"] = function()
        before = function()
            program1 = Mock:new()
            portA:transferProgram( program1 )
            packet = portA:getPacketWithProgram( program1 )
            portA:deliverPacket( packet )
        end
        
        it["should remove the packet"] = function()
            expect( portA:hasPacket( packet ) ).should_be( false )
        end
        
        it["should add the programs inside the packet to the sister sytem"] = function()
            local calls = #Mock.calls[ MockSystemControllerB.addProgram ]
            local call_1_argument_2 = Mock.calls[ MockSystemControllerB.addProgram ][1][2]
            expect( calls ).should_be( 1 )
            expect( call_1_argument_2 ).should_be( program1 )
        end
    end
    
    describe["port forwarding"] = function()
        before = function()
            portC = PortController( MockSystemControllerA, DIR_NE, MockPortRenderer ) 
        end
        
       describe["when forwarding to a connected port in the same system"] = function()
           before = function()
               portC:connectToPort( Mock:new() )
               success = portA:forwardToPort( portC )
           end

           it["should succeed"] = function()
               expect( success ).should_be( true )
           end

           it["should have an inverse relationship with the port"] = function()
               expect( portA.forwardPort ).should_be( portC )
               expect( portC.forwardPort ).should_be( portA )
           end

           it["should be the master of the relationship"] = function()
               expect( portA:isForwardMaster() ).should_be( true )
               expect( portC:isForwardMaster() ).should_be( false )
           end
           
           it["should be in a 'port forwarded' state"] = function()
                expect( portA.state ).should_be( PORT_FORWARDED )
                expect( portC.state ).should_be( PORT_FORWARDED )
           end
           
           it["should dissallow programs from entering"] = function()
                local program = Mock:new()
                expect( portA:canTransferProgram( program ) ).should_be( false )
           end
       end
       
       describe["when forwarding to an unconnected port in the same system"] = function()
            it["should fail"] = function()
                portC.sisterPort = nil
                success = portA:forwardToPort( portC )
                expect( success ).should_be( false )
            end
       end

       describe["when forwarding to a port in a different system"] = function()
           it["should fail"] = function()
               success = portA:forwardToPort( portB )
               expect( success ).should_be( false )
           end
       end
       
       describe["when forwarding to a connected port that loops back to this"] = function()
            it["should fail"] = function()
                local portB = Mock:new()
                local portC = Mock:new()
                local portD = Mock:new()
                portB.sisterPort  = portC
                portB.isConnected:returns( true )
                portB.systemController = portA.systemController
                portB.isLoopedToPort = PortController.isLoopedToPort
                portC.forwardPort = portD
                portC.isForwarded:returns( true )
                portD.sisterPort  = portA
                
                expect( PortController.isLoopedToPort( portB, portA ) ).should_be( true )
                expect( portA:forwardToPort( portB ) ).should_be( false )
            end
       end
       
       describe["when a packet arrives from a sister port"] = function()
            before = function()
                portC:connectToPort( Mock:new() )
                packet = {}
                portA:forwardToPort( portC )
                portA:onPortDeliveredPacket( portB, packet )
            end
            
            it["should move the packet to the forwarded port"] = function()
                expect( portA.forwardPort:hasPacket( packet ) ).should_be( true )
                expect( portA:hasPacket( packet ) ).should_be( false )
            end
       end
       
       describe["when a packet arrives from a forward port"] = function()
            before = function()
                mockPort = Mock:new()
                packet = { transferProgress = 100 }
                portA:connectToPort( mockPort )
                portA:onPortForwardedPacket( Mock:new(), packet )
            end
            
            it["should start transfering packet to the sister port"] = function()
                expect( Mock.getNumberOfCalls( mockPort.onPortTransferPacket ) ).should_be( 1 )
                expect( Mock.getArgumentFromCall( mockPort.onPortTransferPacket, 1, 3 ) ).should_be( packet )
            end
            
            it["should reset the packet's transfer progress to 0"] = function()
                expect( packet.transferProgress ).should_be( 0 )
            end
       end
    end
end